"""
Problem setup for HT fenics-fenics tutorial
"""

from dolfin import *
from fenics import *
from my_enums import DomainPart, ProblemType
import numpy as np
from mshr import *


class ExcludeStraightBoundary(SubDomain):
    def get_user_input_args(self, args):
        self._interface = args.interface

    def inside(self, x, on_boundary):
        tol = 1E-14
        if on_boundary and not near(x[0], x_coupling, tol) or near(x[1], y_top, tol) or near(x[1], y_bottom, tol):
            return True
        else:
            return False


class ExcludeCircleBoundary(SubDomain):
    def inside(self, x, on_boundary):
      tol = 1e-16
      return (near(x[1],0, 1e-16) and x[0]>=0 and x[0]<=x13+0.5*l) or (near(x[1],0, 1e-16) and x[0]>=x16-0.5*l and x[0]<=Lx+0.5*l)  or (near(x[0],x13, 1e-16) and x[1]>=0 and x[1]<=x12 ) or (near(x[0],x16, 1e-16) and x[1]>=0 and x[1]<=x12 )  or near(x[1], Ly, tol) or near(x[0], Lx, tol)  or near(x[0], 0,tol)


class StraightBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = 1E-14
        if on_boundary and near(x[0], x_coupling, tol):
            return True
        else:
            return False


class CircleBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = 1E-14
        point = Point(x[0], x[1])
        if on_boundary and point.distance(midpoint)**2 - radius**2 < tol:
            return True
        else:
            return False

class CavityinterfaceBoundary(SubDomain):
    def inside(self, x, on_boundary):

               
      return (near(x[1],x12, 1e-16) and x[0]>= x19 and x[0]<=(x16+0.5*l) ) or (near(x[1],x12, 1e-16) and x[0]>=(x13-0.5*l) and x[0]<=(x11+0.5*l) ) or (near(x[1],x14, 1e-16) and x[0]>=x20-0.5*l and x[0]<=x18+0.5*l )    or (near(x[0],x11, 1e-16) and x[1]>=x14-0.5*l and x[1]<=x12+0.5*l )   or  (near(x[0],x19, 1e-16) and x[1]>=x14-0.5*l and x[1]<=x12+0.5*l ) 
#**************************************************************
l=1e-6# the interface thickness     
nx=200  # number of grid
ny=200
nnx = nx
nny = ny
Lx=nx*l
Ly=ny*l
#*****************************************************************
low_resolution = nx*0.4   # solid
high_resolution = nx*0.8   # fluid 
#***************************************************************
y_bottom, y_top = 0, Ly
x_left, x_right = 0, Lx
y_coupling = 0.25*Ly  # x coordinate of coupling interface
midpoint = Point(0.5*Lx, 0.5*Ly)

#*****************************************************************
x12 = 28*l # 8*l
x13 = Lx/2-2*x12 # 15*l
x16 = Lx/2+2*x12
x11 = Lx/2-x12/4 # 3*l
x14 = 0.5*x12
x20 = x11
x19 = x11+x14
x18 = x11+x14
x17 = x14
x15 = 0

def get_problem_setup(args):
    if args.dirichlet and not args.neumann:
        problem = ProblemType.DIRICHLET
    elif args.neumann and not args.dirichlet:
        problem = ProblemType.NEUMANN
    elif args.dirichlet and args.neumann:
        raise Exception("you can only choose either a dirichlet problem (option -d) or a neumann problem (option -n)")
    elif not (args.dirichlet or args.neumann):
        raise Exception("you have to choose either a dirichlet problem (option -d) or a neumann problem (option -n)")
    else:
        raise Exception("invalid control flow.")

    if args.interface == 'simple':
        if args.domain == 'circular' or args.domain == 'rectangle':
            raise Exception("Only --domain left or --domain right is supported for --interface {}. "
                            "Invalid --domain {} provided.".format(args.interface, args.domain))
        elif args.domain == 'left':
            return DomainPart.LEFT, problem
        elif args.domain == 'right':
            return DomainPart.RIGHT, problem
        elif args.dirichlet and args.neumann:
            raise Exception("You can only choose to either compute the left part of the domain (option -dl) "
                            "or the right part (option -dr)")
        elif not (args.domain == 'left' or args.domain == 'right'):
            print("Default domain partitioning for simple interface is used: Left part of domain is a "
                  "Dirichlet-type problem; right part is a Neumann-type problem")
            if problem is ProblemType.DIRICHLET:
                return DomainPart.LEFT, problem
            elif problem is ProblemType.NEUMANN:
                return DomainPart.RIGHT, problem
    elif args.interface == 'complex':
        if args.domain == 'left' or args.domain == 'right':
            raise Exception("Only --domain circular or --domain rectangle is supported for --interface {}. "
                            "Invalid --domain {} provided.".format(args.interface, args.domain))
        elif args.domain == 'circular':
            return DomainPart.CIRCULAR, problem
        elif args.domain == 'rectangle':
            return DomainPart.RECTANGLE, problem
        elif args.dirichlet and args.neumann:
            raise Exception("You can only choose to either compute the circular part of the domain (option -dc) "
                            "or the residual part (option -dnc)")
        elif not (args.domain == 'circular' or args.domain == 'rectangle'):
            print("Default domain partitioning for complex interface is used: Circular part of domain "
                  "is a Neumann-type problem; Rest of the domain is a Dirichlet-type problem")
            if problem is ProblemType.NEUMANN:
                return DomainPart.CIRCULAR, problem
            elif problem is ProblemType.DIRICHLET:
                return DomainPart.RECTANGLE, problem
    else:
        raise Exception("invalid interface provided: args.interface = {}".format(args.interface))


def get_geometry(domain_part):
    nx = 5
    ny = 10
    n_vertices = 20

    if domain_part is DomainPart.LEFT:
        nx = nx * 3
    elif domain_part is DomainPart.RIGHT:
        ny = ny * 2
    elif domain_part is DomainPart.CIRCULAR:
        n_vertices = n_vertices
    elif domain_part is DomainPart.RECTANGLE:
        n_vertices = n_vertices
    else:
        raise Exception("invalid domain_part: {}".format(domain_part))

    if domain_part is DomainPart.LEFT or domain_part is DomainPart.RIGHT:
        if domain_part is DomainPart.LEFT:
            p0 = Point(x_left, y_bottom)
            p1 = Point(x_coupling, y_top)
        elif domain_part is DomainPart.RIGHT:
            p0 = Point(x_coupling, y_bottom)
            p1 = Point(x_right, y_top)
        else:
            raise Exception("invalid control flow!")
        mesh = RectangleMesh(p0, p1, nx, ny)
        coupling_boundary = StraightBoundary()
        remaining_boundary = ExcludeStraightBoundary()

    elif domain_part is DomainPart.CIRCULAR or domain_part is DomainPart.RECTANGLE:
        p0 = Point(x_left, y_bottom)
        p1 = Point(x_right, y_top)
        whole_domain = Rectangle(p0, p1)
        if domain_part is DomainPart.CIRCULAR:
            cavity_domain = Polygon([Point(0,0),Point(x13,0),Point(x13,x12),
                                    Point(x11,x12),Point(x20,x14),Point(x18,x17),
                                    Point(x19,x12), Point(x16,x12),Point(x16,0), Point(Lx,0),Point(Lx,Ly),Point(0,Ly)])

            mesh = generate_mesh(cavity_domain, high_resolution, "cgal")
            mesh_file = File('out/fluidmesh/fluidmesh.pvd')
            mesh_file<<mesh
            boundaries = MeshFunction("size_t", mesh, 1)
            boundaries.set_all(0)
            coupling_boundary = CavityinterfaceBoundary()
            remaining_boundary = ExcludeCircleBoundary()
            coupling_boundary.mark(boundaries,10)
            remaining_boundary.mark(boundaries,20)
            File('out/boundarymark/facetfunc.pvd').write(boundaries)
        elif domain_part is DomainPart.RECTANGLE:
            cavity_domain = Polygon([Point(0,0),Point(x13,0),Point(x13,x12),
                                    Point(x11,x12),Point(x20,x14),Point(x18,x17),
                                    Point(x19,x12), Point(x16,x12),Point(x16,0), Point(Lx,0),Point(Lx,Ly),Point(0,Ly)])                                
            mesh = generate_mesh(whole_domain - cavity_domain , low_resolution, "cgal")
            mesh_file = File('out/solidmesh/solidmesh.pvd')
            mesh_file<<mesh
        else:
            raise Exception("invalid control flow!")
        coupling_boundary = CavityinterfaceBoundary()
        remaining_boundary = ExcludeCircleBoundary()

    else:
        raise Exception("invalid control flow!")

    return mesh, coupling_boundary, remaining_boundary
